import { Observable, ObservableInput } from 'rxjs';
import type { MonoTypeOperatorFunction } from 'rxjs';
import { AsyncValue } from './AsyncValue';
/**
 * VoAV stands for "Value or AsyncValue".
 * i.e. VoAV<number> is used where both, number and AsyncValue<number> are acceptable
 */
export declare type VoAV<T> = T | AsyncValue<T>;
/**
 * VoAV stands for "Value or AsyncValue".
 * PromiseLikeVoAV<T> is PromiseLike<VoAV<T>>
 * i.e. Promise<VoAV<number>> is used where both, PromiseLike<number> and PromiseLike<AsyncValue<number>> are acceptable
 */
export declare type PromiseLikeVoAV<T> = PromiseLike<VoAV<T>>;
/**
 * VoAV stands for "Value or AsyncValue".
 * ObservableVoAV<T> is Observable<VoAV<T>>
 * i.e. Promise<VoAV<number>> is used where both, Observable<number> and Observable<AsyncValue<number>> are acceptable
 */
export declare type ObservableVoAV<T> = Observable<VoAV<T>>;
/**
 * VoAV stands for "Value or AsyncValue".
 * AsyncLikeVoAV<T> = VoAV<T> | PromiseLikeVoAV<T>
 */
export declare type AsyncLikeVoAV<T> = VoAV<T> | ObservableVoAV<T> | ObservableVoAV<T>;
/**
 * returns true when value is not of type AsyncValue
 * or is an AsyncFalue and it is fulfilled
 */
export declare function isFulfilledOrSync(value: unknown): boolean;
export declare function getSyncValue<T>(voAv: VoAV<T>): T;
/**
 * async values are equal if their errors are equal '==='
 * or their pendings are equal
 * or their values are equal '==='
 *
 * if @compare function is provided, it
 * it will be used to compare value properites instead of ===
 */
export declare function compareAsyncValues<T>(a1: AsyncValue<T>, a2: AsyncValue<T>, compare?: (v1: T, v2: T) => boolean): boolean;
export declare function convertToAsyncValue<T>(value: T | AsyncValue<T>): AsyncValue<T>;
/**
 *
 * @returns Observable from promise
 * Observable will emit pending state right away
 * and AsyncValue<sucessfull> when promise resolves or AsyncValue<error> when Promise rejects
 *
 * // marble diagram (p: Pending, v: Value, e: Error):
 * 'p---v' | 'p---e'
 */
export declare function promiseToAsyncValueStream<T>(promise: PromiseLike<T | AsyncValue<T>>): Observable<AsyncValue<T>>;
/**
 * this is an RXJS Operator
 *
 * maps all stream elements of type T to AsyncValue<T>
 * if stream emits an error, AsyncValue.errorOnly(error) will be emitted
 */
export declare function mapToAsyncValue<T>(): (source: Observable<VoAV<T>>) => Observable<AsyncValue<T>>;
/**
 * this is an RXJS Operator
 *
 * prevents emitting of duplicate AsyncValues
 * async values are duplicates if their errors are equal '==='
 * or their pendings are equal
 * or their values are equal '==='
 *
 * if @compare function is provided, it
 * it will be used to compare value properites instead of ===
 */
export declare function distinctAsyncValueUntilChanged<T>(compare?: (v1: T, v2: T) => boolean): MonoTypeOperatorFunction<AsyncValue<T>>;
/**
 * if any of @sources is an AsyncValue containing error,
 * let's call it sourceWithError,
 * sourceWithError will be returned
 *
 * otherwise if any of @sources is an AsyncValue in pending state
 * new AsyncValue.pendingOnly() will be returned
 *
 * otherwise, when fulfilled, if @reducerOrResult is a function
 * @reducerOrResult(sources) will be returned
 *
 * otherwise @reducerOrResult will be returned
 */
export declare function internalReduceAsyncValues<T, R>(sources: any[], reducerOrResult: ((sourceValues: T[]) => R) | AsyncValue<R> | R): AsyncValue<R> | R;
/**
 * if any of @sources is an AsyncValue containing error,
 * let's call it sourceWithError,
 * sourceWithError will be returned
 *
 * otherwise if any of @sources is an AsyncValue in pending state
 * new AsyncValue.pendingOnly() will be returned
 *
 * otherwise, when fulfilled, if @reducerOrResult is a function
 * @reducerOrResult(sources), converted to AsyncValue if already not one, will be returned
 *
 * otherwise @reducerOrResult, converted to AsyncValue if already not one, will be returned
 */
export declare function reduceAsyncValues<T, R>(sources: any[], reducerOrResult: ((sourceValues: T[]) => R) | VoAV<R>): AsyncValue<R>;
/**
 * this is an RXJS Operator
 */
export declare function mapToWhenFulfilled<T, R>(outputValue: VoAV<R>): (source: Observable<VoAV<T>>) => Observable<AsyncValue<R>>;
/**
 * this is an RXJS Operator
 *
 * if @source is an AsyncValue containing error
 * it's clone will be emitted
 *
 * otherwise if @source is a pending AsyncValue
 * new AsyncValue.pendingOnly() will be emitted
 *
 * otherwise, when fulfilled, if @transformFnOrResult is a function
 * @combinerOrResult(sources), converted to AsyncValue if already not one, will be emitted
 *
 * otherwise @combinerOrResult, converted to AsyncValue if already not one, will be emitted
 */
export declare function mapWhenFulfilled<T, R>(transformFnOrResult: VoAV<R> | ((sourceValue: T, index?: number) => R)): (source: Observable<AsyncValue<T>>) => Observable<AsyncValue<R>>;
/**
 * this is an RXJS Operator
 *
 * if @source is in error or pending state, clone of the AsyncValue will be emitted
 *
 * otherwise if @transformFn is a Function
 * @transformFn (AsyncValue<fulfilled>) will became new inner observable
 * and every value it emits will be mapped to an AsyncValue
 */
export declare function switchMapWhenFulfilled<T, R>(transformFnOrValue: ((sourceValue: T) => VoAV<R> | PromiseLikeVoAV<R> | ObservableVoAV<R>) | VoAV<R> | PromiseLikeVoAV<R> | ObservableVoAV<R>): (source: Observable<VoAV<T>>) => Observable<AsyncValue<R>>;
export declare function combineLatestWhenAllFulfilled<R, T1>([o1]: [ObservableInput<VoAV<T1>>], combinerOrResult: ((sourceValues: [T1]) => VoAV<R> | PromiseLikeVoAV<R> | ObservableVoAV<R>) | VoAV<R> | PromiseLikeVoAV<R> | ObservableVoAV<R>): Observable<AsyncValue<R>>;
export declare function combineLatestWhenAllFulfilled<R, T1, T2>(sources: [ObservableInput<VoAV<T1>>, ObservableInput<VoAV<T2>>], combinerOrResult: ((sourceValues: [T1, T2]) => VoAV<R> | PromiseLikeVoAV<R> | ObservableVoAV<R>) | VoAV<R> | PromiseLikeVoAV<R> | ObservableVoAV<R>): Observable<AsyncValue<R>>;
export declare function combineLatestWhenAllFulfilled<R, T1, T2, T3>(sources: [ObservableInput<VoAV<T1>>, ObservableInput<VoAV<T2>>, ObservableInput<VoAV<T3>>], combinerOrResult: ((sourceValues: [T1, T2, T3]) => VoAV<R> | PromiseLikeVoAV<R> | ObservableVoAV<R>) | VoAV<R> | PromiseLikeVoAV<R> | ObservableVoAV<R>): Observable<AsyncValue<R>>;
export declare function combineLatestWhenAllFulfilled<R, T1, T2, T3, T4>(sources: [ObservableInput<VoAV<T1>>, ObservableInput<VoAV<T2>>, ObservableInput<VoAV<T3>>, ObservableInput<VoAV<T4>>], combinerOrResult: ((sourceValues: [T1, T2, T3, T4]) => VoAV<R> | PromiseLikeVoAV<R> | ObservableVoAV<R>) | VoAV<R> | PromiseLikeVoAV<R> | ObservableVoAV<R>): Observable<AsyncValue<R>>;
export declare function combineLatestWhenAllFulfilled<R, T1, T2, T3, T4, T5>(sources: [ObservableInput<VoAV<T1>>, ObservableInput<VoAV<T2>>, ObservableInput<VoAV<T3>>, ObservableInput<VoAV<T4>>, ObservableInput<VoAV<T5>>], combinerOrResult: ((sourceValues: [T1, T2, T3, T4, T5]) => VoAV<R> | PromiseLikeVoAV<R> | ObservableVoAV<R>) | VoAV<R> | PromiseLikeVoAV<R> | ObservableVoAV<R>): Observable<AsyncValue<R>>;
export declare function combineLatestWhenAllFulfilled<R, T1, T2, T3, T4, T5, T6>(sources: [ObservableInput<VoAV<T1>>, ObservableInput<VoAV<T2>>, ObservableInput<VoAV<T3>>, ObservableInput<VoAV<T4>>, ObservableInput<VoAV<T5>>, ObservableInput<VoAV<T6>>], combinerOrResult: ((sourceValues: [T1, T2, T3, T4, T5, T6]) => VoAV<R> | PromiseLikeVoAV<R> | ObservableVoAV<R>) | VoAV<R> | PromiseLikeVoAV<R> | ObservableVoAV<R>): Observable<AsyncValue<R>>;
export declare function combineLatestWhenAllFulfilled<R, T1, T2, T3, T4, T5, T6, T7>(sources: [ObservableInput<VoAV<T1>>, ObservableInput<VoAV<T2>>, ObservableInput<VoAV<T3>>, ObservableInput<VoAV<T4>>, ObservableInput<VoAV<T5>>, ObservableInput<VoAV<T6>>, ObservableInput<VoAV<T7>>], combinerOrResult: ((sourceValues: [T1, T2, T3, T4, T5, T6, T7]) => VoAV<R> | PromiseLikeVoAV<R> | ObservableVoAV<R>) | VoAV<R> | PromiseLikeVoAV<R> | ObservableVoAV<R>): Observable<AsyncValue<R>>;
export declare function combineLatestWhenAllFulfilled<R, T1, T2, T3, T4, T5, T6, T7, T8>(sources: [ObservableInput<VoAV<T1>>, ObservableInput<VoAV<T2>>, ObservableInput<VoAV<T3>>, ObservableInput<VoAV<T4>>, ObservableInput<VoAV<T5>>, ObservableInput<VoAV<T6>>, ObservableInput<VoAV<T7>>, ObservableInput<VoAV<T8>>], combinerOrResult: ((sourceValues: [T1, T2, T3, T4, T5, T6, T7, T8]) => VoAV<R> | PromiseLikeVoAV<R> | ObservableVoAV<R>) | VoAV<R> | PromiseLikeVoAV<R> | ObservableVoAV<R>): Observable<AsyncValue<R>>;
export declare function combineLatestWhenAllFulfilled<R, T1, T2, T3, T4, T5, T6, T7, T8, T9>(sources: [ObservableInput<VoAV<T1>>, ObservableInput<VoAV<T2>>, ObservableInput<VoAV<T3>>, ObservableInput<VoAV<T4>>, ObservableInput<VoAV<T5>>, ObservableInput<VoAV<T6>>, ObservableInput<VoAV<T7>>, ObservableInput<VoAV<T8>>, ObservableInput<VoAV<T9>>], combinerOrResult: ((sourceValues: [T1, T2, T3, T4, T5, T6, T7, T8, T9]) => VoAV<R> | PromiseLikeVoAV<R> | ObservableVoAV<R>) | VoAV<R> | PromiseLikeVoAV<R> | ObservableVoAV<R>): Observable<AsyncValue<R>>;
/**
 * this is an RXJS Operator
 */
export declare function transformWhenAllFulfilled<T, R>(transformFnOrResult: ((sourceValues: T[]) => R) | VoAV<R>): (sources: Observable<AsyncValue<T>>[]) => Observable<AsyncValue<R>>;
